/*=======================================================================
              bpg load plugin for the GIMP
                 Copyright (C) 2014 - Mikro

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
=======================================================================*/

#include <stdlib.h>
#include <stdint.h>
#include <libgimp/gimp.h>

#include "libbpg.h"

#include "read-bpg.h"

int read_bpg(const gchar * filename)
{
    FILE * file;
    buffer buf;
    
    gint32 new_image_id, new_layer_id;
    GimpDrawable * drawable;
    GimpPixelRgn region;
    GimpImageBaseType image_base_type;
    
    BPGDecoderContext *image_context;
    BPGImageInfo image_info_s, *image_info = &image_info_s;
    BPGDecoderOutputFormat output_format;
    
    int bit_depth = 8;
    int y;
    int bpp;
    guchar *image_data;
    
    // open the file
    file = fopen(filename, "rb");
    
    if(!file) return -1;

    // get the file size
    fseek(file, 0, SEEK_END);
    buf.length = ftell(file);
    fseek(file, 0, SEEK_SET);
    
    // read the file
    buf.data = malloc(buf.length);
    if (fread(buf.data, 1, buf.length, file) != buf.length) {
        fprintf(stderr, "Error while reading bpg file\n");
        return -1;
    }
    
    // close the file
    fclose(file);

    // create the image context
    image_context = bpg_decoder_open();
    
    // decode the bpg
    if (bpg_decoder_decode(image_context, buf.data, buf.length) < 0) {
        fprintf(stderr, "Could not decode image\n");
        return -1;
    }
    
    free(buf.data);
    
    // get some info
    bpg_decoder_get_info(image_context, image_info);
    
    // set the output format
    if (bit_depth == 16) {
        if (image_info->has_alpha)
            output_format = BPG_OUTPUT_FORMAT_RGBA64;
        else
            output_format = BPG_OUTPUT_FORMAT_RGB48;
    } else {
        if (image_info->has_alpha)
            output_format = BPG_OUTPUT_FORMAT_RGBA32;
        else
            output_format = BPG_OUTPUT_FORMAT_RGB24;
    }

    if (image_info->has_alpha) {
      image_base_type = GIMP_RGBA_IMAGE;
    } else {
      image_base_type = GIMP_RGB_IMAGE;
    }
    
    // create a new canvas
    new_image_id = gimp_image_new(image_info->width, image_info->height, GIMP_RGB);
    
    // create an empty layer
    new_layer_id = gimp_layer_new(new_image_id,
                                  "background",
                                  image_info->width, image_info->height,
                                  image_base_type,
                                  100,
                                  GIMP_NORMAL_MODE);

    // get the drawable for this layer
    drawable = gimp_drawable_get(new_layer_id);

    // obtain the full region of this layer. Raw pixel info goes here
    gimp_pixel_rgn_init(&region,
                        drawable,
                        0, 0,
                        image_info->width, image_info->height,
                        TRUE, FALSE);

    // start data decoder
    bpg_decoder_start(image_context, output_format);
    
    bpp = (3 + image_info->has_alpha) * (bit_depth / 8);

    // a buffer to temporally store the info
    image_data = (guchar *)calloc(sizeof(guchar *),image_info->width * image_info->height);

    // put into the buffer the image data, one pixel line at a time
    for (y = 0; y < image_info->height; y++) {
      bpg_decoder_get_line( image_context, image_data + (y * image_info->width * bpp) ); 
    }
    
    // copy the buffer to the drawable
    gimp_pixel_rgn_set_rect(&region,
			    image_data,
			    0, 0,
			    image_info->width, image_info->height);

    // cleanup
    free(image_data);
    gimp_drawable_flush(drawable);
    gimp_drawable_detach(drawable);
    
    
    // add the layer to the image
    gimp_image_add_layer(new_image_id, new_layer_id, 0);
    
    // return the new image id
    return new_image_id;
}
